@@ -1,9 +1,3 @@
-# -- don't import anything here that could import Theano
-#    Because theano configures itself on import.... we want to leave
-#    room for the Bandit being loaded by mongo-worker to set things up
-#    for itself.
-
-
 from .base import STATUS_STRINGS
 from .base import STATUS_NEW
 from .base import STATUS_RUNNING
@@ -17,33 +11,31 @@ from .base import JOB_STATE_RUNNING
 from .base import JOB_STATE_DONE
 from .base import JOB_STATE_ERROR
 
-from .base import Bandit
-from .base import as_bandit
-from .base import BanditAlgo
 from .base import Ctrl
-from .base import Experiment
-from .base import StopExperiment
-from .base import Random
 from .base import Trials
 from .base import trials_from_docs
+from .base import Domain
 
 from .fmin import fmin
 from .fmin import fmin_pass_expr_memo_ctrl
-from .fmin import Domain
 from .fmin import FMinIter
 from .fmin import partial
 from .fmin import space_eval
 
 # -- syntactic sugar
-import hp
+from . import hp
 
 # -- exceptions
-import exceptions
+from . import exceptions
 
 # -- Import built-in optimization algorithms
-import rand
-import tpe
+from . import rand
+from . import tpe
+from . import atpe
+from . import mix
+from . import anneal
 
-# -- TODO: deprecate in favour of tpe.suggest
-from tpe import TreeParzenEstimator
+# -- spark extension
+from .spark import SparkTrials
 
+__version__ = "0.2.6"
