@@ -2,34 +2,37 @@
 #
 # It provides types to build ASTs in a simple lambda-notation style
 #
-
+from future import standard_library
 import copy
-import logging; logger = logging.getLogger(__name__)
+import logging
+import operator
 import time
 
-from StringIO import StringIO
 from collections import deque
 
 import networkx as nx
 
 # TODO: move things depending on numpy (among others too) to a library file
 import numpy as np
-np_versions = map(int, np.__version__.split('.')[:2])
+import six
+from six import StringIO
 
+standard_library.install_aliases()
+logger = logging.getLogger(__name__)
+np_versions = list(map(int, np.__version__.split(".")[:2]))
 
 DEFAULT_MAX_PROGRAM_LEN = 100000
 
 
 class PyllImportError(ImportError):
-    """A pyll symbol was not defined in the scope """
+    """A pyll symbol was not defined in the scope"""
 
 
-class MissingArgument(object):
-    """Object to represent a missing argument to a function application
-    """
+class MissingArgument:
+    """Object to represent a missing argument to a function application"""
 
 
-class SymbolTable(object):
+class SymbolTable:
     """
     An object whose methods generally allocate Apply nodes.
 
@@ -42,100 +45,111 @@ class SymbolTable(object):
     def __init__(self):
         # -- list and dict are special because they are Python builtins
         self._impls = {
-                'list': list,
-                'dict': dict,
-                'range': range,
-                'len': len,
-                'int': int,
-                'float': float,
-                'map': map,
-                'max': max,
-                'min': min,
-                'getattr': getattr,
-                }
+            "list": list,
+            "dict": dict,
+            "range": range,
+            "len": len,
+            "int": int,
+            "float": float,
+            "map": map,
+            "max": max,
+            "min": min,
+            "getattr": getattr,
+        }
 
     def _new_apply(self, name, args, kwargs, o_len, pure):
         pos_args = [as_apply(a) for a in args]
-        named_args = [(k, as_apply(v)) for (k, v) in kwargs.items()]
+        named_args = [(k, as_apply(v)) for (k, v) in list(kwargs.items())]
         named_args.sort()
-        return Apply(name,
-                pos_args=pos_args,
-                named_args=named_args,
-                o_len=o_len,
-                pure=pure)
-
-    # ----
+        return Apply(
+            name, pos_args=pos_args, named_args=named_args, o_len=o_len, pure=pure
+        )
 
     def dict(self, *args, **kwargs):
         # XXX: figure out len
-        return self._new_apply('dict', args, kwargs, o_len=None,
-                pure=True)
+        return self._new_apply("dict", args, kwargs, o_len=None, pure=True)
 
     def int(self, arg):
-        return self._new_apply('int', [as_apply(arg)], {}, o_len=None,
-                pure=True)
+        return self._new_apply("int", [as_apply(arg)], {}, o_len=None, pure=True)
 
     def float(self, arg):
-        return self._new_apply('float', [as_apply(arg)], {}, o_len=None,
-                pure=True)
+        return self._new_apply("float", [as_apply(arg)], {}, o_len=None, pure=True)
 
     def len(self, obj):
-        return self._new_apply('len', [obj], {}, o_len=None,
-                pure=True)
+        return self._new_apply("len", [obj], {}, o_len=None, pure=True)
 
     def list(self, init):
-        return self._new_apply('list', [as_apply(init)], {}, o_len=None,
-                pure=True)
+        return self._new_apply("list", [as_apply(init)], {}, o_len=None, pure=True)
 
     def map(self, fn, seq, pure=False):
         """
         pure - True is assertion that fn does not modify seq[i]
         """
-        return self._new_apply('map', [as_apply(fn), as_apply(seq)], {},
-                               o_len=seq.o_len,
-                               pure=pure
-                               )
+        return self._new_apply(
+            "map", [as_apply(fn), as_apply(seq)], {}, o_len=seq.o_len, pure=pure
+        )
 
     def range(self, *args):
-        return self._new_apply('range', args, {}, o_len=None, pure=True)
+        return self._new_apply("range", args, {}, o_len=None, pure=True)
 
     def max(self, *args):
-        """ return max of args """
-        return self._new_apply('max', map(as_apply, args), {},
-                o_len=None, pure=True)
+        """return max of args"""
+        return self._new_apply(
+            "max", list(map(as_apply, args)), {}, o_len=None, pure=True
+        )
 
     def min(self, *args):
-        """ return min of args """
-        return self._new_apply('min', map(as_apply, args), {},
-                o_len=None, pure=True)
+        """return min of args"""
+        return self._new_apply(
+            "min", list(map(as_apply, args)), {}, o_len=None, pure=True
+        )
 
     def getattr(self, obj, attr, *args):
-        return self._new_apply('getattr',
-                [as_apply(obj), as_apply(attr)] + map(as_apply, args),
-                {},
-                o_len=None,
-                pure=True)
+        return self._new_apply(
+            "getattr",
+            [as_apply(obj), as_apply(attr)] + list(map(as_apply, args)),
+            {},
+            o_len=None,
+            pure=True,
+        )
 
-    # ----
-
-    def define(self, f, o_len=None, pure=False):
-        """Decorator for adding python functions to self
-        """
+    def _define(self, f, o_len, pure):
         name = f.__name__
-        if hasattr(self, name):
-            raise ValueError('Cannot override existing symbol', name)
-
         entry = SymbolTableEntry(self, name, o_len, pure)
         setattr(self, name, entry)
         self._impls[name] = f
         return f
 
+    def define(self, f, o_len=None, pure=False):
+        """Decorator for adding python functions to self"""
+        name = f.__name__
+        if hasattr(self, name):
+            raise ValueError("Cannot override existing symbol", name)
+        return self._define(f, o_len, pure)
+
+    def define_if_new(self, f, o_len=None, pure=False):
+        """Pass silently if f matches the current implementation
+        for f.__name__"""
+        name = f.__name__
+        if hasattr(self, name) and self._impls[name] is not f:
+            raise ValueError("Cannot redefine existing symbol", name)
+        return self._define(f, o_len, pure)
+
+    def undefine(self, f):
+        if isinstance(f, str):
+            name = f
+        else:
+            name = f.__name__
+        del self._impls[name]
+        delattr(self, name)
+
     def define_pure(self, f):
         return self.define(f, o_len=None, pure=True)
 
     def define_info(self, o_len=None, pure=False):
         def wrapper(f):
             return self.define(f, o_len=o_len, pure=pure)
+
         return wrapper
 
     def inject(self, *args, **kwargs):
@@ -150,7 +164,7 @@ class SymbolTable(object):
                 rval[k] = getattr(self, k)
             except AttributeError:
                 raise PyllImportError(k)
-        for k, origk in kwargs.items():
+        for k, origk in list(kwargs.items()):
             try:
                 rval[k] = getattr(self, origk)
             except AttributeError:
@@ -160,9 +174,10 @@ class SymbolTable(object):
     def import_(self, _globals, *args, **kwargs):
         _globals.update(self.inject(*args, **kwargs))
 
-class SymbolTableEntry(object):
-    """A functools.partial-like class for adding symbol table entries.
-    """
+
+class SymbolTableEntry:
+    """A functools.partial-like class for adding symbol table entries."""
+
     def __init__(self, symbol_table, apply_name, o_len, pure):
         self.symbol_table = symbol_table
         self.apply_name = apply_name
@@ -171,43 +186,40 @@ class SymbolTableEntry(object):
 
     def __call__(self, *args, **kwargs):
         return self.symbol_table._new_apply(
-                self.apply_name,
-                args,
-                kwargs,
-                self.o_len,
-                self.pure)
+            self.apply_name, args, kwargs, self.o_len, self.pure
+        )
+
 
 scope = SymbolTable()
 
 
 def as_apply(obj):
-    """Smart way of turning object into an Apply
-    """
+    """Smart way of turning object into an Apply"""
     if isinstance(obj, Apply):
         rval = obj
     elif isinstance(obj, tuple):
-        rval = Apply('pos_args', [as_apply(a) for a in obj], {}, len(obj))
+        rval = Apply("pos_args", [as_apply(a) for a in obj], {}, len(obj))
     elif isinstance(obj, list):
-        rval = Apply('pos_args', [as_apply(a) for a in obj], {}, None)
+        rval = Apply("pos_args", [as_apply(a) for a in obj], {}, None)
     elif isinstance(obj, dict):
-        items = obj.items()
+        items = list(obj.items())
         # -- should be fine to allow numbers and simple things
-        #    but think about if it's ok to allow Applys
+        #    but think about if it's ok to allow Apply objects
         #    it messes up sorting at the very least.
         items.sort()
-        if all(isinstance(k, basestring) for k in obj):
+        if all(isinstance(k, str) for k in obj):
             named_args = [(k, as_apply(v)) for (k, v) in items]
-            rval = Apply('dict', [], named_args, len(named_args))
+            rval = Apply("dict", [], named_args, len(named_args))
         else:
             new_items = [(k, as_apply(v)) for (k, v) in items]
-            rval = Apply('dict', [as_apply(new_items)], {}, o_len=None)
+            rval = Apply("dict", [as_apply(new_items)], {}, o_len=None)
     else:
         rval = Literal(obj)
     assert isinstance(rval, Apply)
     return rval
 
 
-class Apply(object):
+class Apply:
     """
     Represent a symbolic application of a symbol to arguments.
 
@@ -216,20 +228,29 @@ class Apply(object):
     pure - True only if the function has no relevant side-effects
     """
 
-    def __init__(self, name, pos_args, named_args,
-            o_len=None,
-            pure=False):
+    def __init__(
+        self, name, pos_args, named_args, o_len=None, pure=False, define_params=None
+    ):
         self.name = name
         # -- tuples or arrays -> lists
         self.pos_args = list(pos_args)
         self.named_args = [[kw, arg] for (kw, arg) in named_args]
         # -- o_len is attached this early to support tuple unpacking and
-        #    list coersion.
+        #    list coercion.
         self.o_len = o_len
         self.pure = pure
+        # -- define_params lets us cope with stuff that may be in the
+        #    SymbolTable on the master but not on the worker.
+        self.define_params = define_params
         assert all(isinstance(v, Apply) for v in pos_args)
         assert all(isinstance(v, Apply) for k, v in named_args)
-        assert all(isinstance(k, basestring) for k, v in named_args)
+        assert all(isinstance(k, str) for k, v in named_args)
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        # -- On deserialization, update scope if need be.
+        if self.define_params:
+            scope.define_if_new(**self.define_params)
 
     def eval(self, memo=None):
         """
@@ -257,14 +278,18 @@ class Apply(object):
             return memo[id(self)]
         else:
             args = [a.eval() for a in self.pos_args]
-            kwargs = dict([(n, a.eval()) for (n, a) in self.named_args])
+            kwargs = {n: a.eval() for (n, a) in self.named_args}
             f = scope._impls[self.name]
             memo[id(self)] = rval = f(*args, **kwargs)
             return rval
 
     def inputs(self):
-        rval = self.pos_args + [v for (k, v) in self.named_args]
-        assert all(isinstance(arg, Apply) for arg in rval)
+        # -- this function gets called a lot and it's not 100% safe to cache
+        #    so the if/else is a small optimization
+        if self.named_args:
+            rval = self.pos_args + [v for (k, v) in self.named_args]
+        else:
+            rval = self.pos_args
         return rval
 
     @property
@@ -293,29 +318,29 @@ class Apply(object):
         try:
             if extra_args_ok and extra_kwargs_ok:
                 assert len(code.co_varnames) >= code.co_argcount + 2
-                param_names = code.co_varnames[:code.co_argcount + 2]
+                param_names = code.co_varnames[: code.co_argcount + 2]
                 args_param = param_names[code.co_argcount]
                 kwargs_param = param_names[code.co_argcount + 1]
-                pos_params = param_names[:code.co_argcount]
+                pos_params = param_names[: code.co_argcount]
             elif extra_kwargs_ok:
                 assert len(code.co_varnames) >= code.co_argcount + 1
-                param_names = code.co_varnames[:code.co_argcount + 1]
+                param_names = code.co_varnames[: code.co_argcount + 1]
                 kwargs_param = param_names[code.co_argcount]
-                pos_params = param_names[:code.co_argcount]
+                pos_params = param_names[: code.co_argcount]
             elif extra_args_ok:
                 assert len(code.co_varnames) >= code.co_argcount + 1
-                param_names = code.co_varnames[:code.co_argcount + 1]
+                param_names = code.co_varnames[: code.co_argcount + 1]
                 args_param = param_names[code.co_argcount]
-                pos_params = param_names[:code.co_argcount]
+                pos_params = param_names[: code.co_argcount]
             else:
                 assert len(code.co_varnames) >= code.co_argcount
-                param_names = code.co_varnames[:code.co_argcount]
-                pos_params = param_names[:code.co_argcount]
+                param_names = code.co_varnames[: code.co_argcount]
+                pos_params = param_names[: code.co_argcount]
         except AssertionError:
-            print 'YIKES: MISUNDERSTANDING OF CALL PROTOCOL:',
-            print code.co_argcount,
-            print code.co_varnames,
-            print '%x' % code.co_flags
+            print("YIKES: MISUNDERSTANDING OF CALL PROTOCOL:")
+            print(code.co_argcount)
+            print(code.co_varnames)
+            print("%x" % code.co_flags)
             raise
 
         if extra_args_ok:
@@ -325,7 +350,7 @@ class Apply(object):
             binding[kwargs_param] == {}
 
         if len(self.pos_args) > code.co_argcount and not extra_args_ok:
-            raise TypeError('Argument count exceeds number of positional params')
+            raise TypeError("Argument count exceeds number of positional params")
 
         # -- bind positional arguments
         for param_i, arg_i in zip(param_names, self.pos_args):
@@ -333,7 +358,8 @@ class Apply(object):
 
         if extra_args_ok:
             # XXX: THIS IS NOT BEING TESTED AND IS OBVIOUSLY BROKEN
-            binding[args_param].extend(args[code.co_argcount:])
+            # TODO: 'args' does not even exist at this point
+            binding[args_param].extend(args[code.co_argcount :])
 
         # -- bind keyword arguments
         for aname, aval in self.named_args:
@@ -344,10 +370,10 @@ class Apply(object):
                     binding[kwargs_param][aname] = aval
                     continue
                 else:
-                    raise TypeError('Unrecognized keyword argument', aname)
+                    raise TypeError("Unrecognized keyword argument", aname)
             param = param_names[pos]
             if param in binding:
-                raise TypeError('Duplicate argument for parameter', param)
+                raise TypeError("Duplicate argument for parameter", param)
             binding[param] = aval
 
         assert len(binding) <= len(param_names)
@@ -366,20 +392,21 @@ class Apply(object):
                 return
         arg = self.arg
         if name in arg and arg[name] != MissingArgument:
-            raise NotImplementedError('change pos arg to kw arg')
+            raise NotImplementedError("change pos arg to kw arg")
         else:
             self.named_args.append([name, as_apply(value)])
             self.named_args.sort()
 
-    def clone_from_inputs(self, inputs, o_len='same'):
+    def clone_from_inputs(self, inputs, o_len="same"):
         if len(inputs) != len(self.inputs()):
             raise TypeError()
         L = len(self.pos_args)
         pos_args = list(inputs[:L])
-        named_args = [[kw, inputs[L + ii]]
-                for ii, (kw, arg) in enumerate(self.named_args)]
+        named_args = [
+            [kw, inputs[L + ii]] for ii, (kw, arg) in enumerate(self.named_args)
+        ]
         # -- danger cloning with new inputs can change the o_len
-        if o_len == 'same':
+        if o_len == "same":
             o_len = self.o_len
         return self.__class__(self.name, pos_args, named_args, o_len)
 
@@ -402,16 +429,16 @@ class Apply(object):
             lineno = [0]
 
         if self in memo:
-            print >> ofile, lineno[0], ' ' * indent + memo[self]
+            print(lineno[0], " " * indent + memo[self], file=ofile)
             lineno[0] += 1
         else:
-            memo[self] = self.name + ('  [line:%i]' % lineno[0])
-            print >> ofile, lineno[0], ' ' * indent + self.name
+            memo[self] = self.name + ("  [line:%i]" % lineno[0])
+            print(lineno[0], " " * indent + self.name, file=ofile)
             lineno[0] += 1
             for arg in self.pos_args:
                 arg.pprint(ofile, lineno, indent + 2, memo)
             for name, arg in self.named_args:
-                print >> ofile, lineno[0], ' ' * indent + ' ' + name + ' ='
+                print(lineno[0], " " * indent + " " + name + " =", file=ofile)
                 lineno[0] += 1
                 arg.pprint(ofile, lineno, indent + 2, memo)
 
@@ -447,6 +474,12 @@ class Apply(object):
     def __rdiv__(self, other):
         return scope.div(other, self)
 
+    def __truediv__(self, other):
+        return scope.truediv(self, other)
+
+    def __rtruediv__(self, other):
+        return scope.truediv(other, self)
+
     def __floordiv__(self, other):
         return scope.floordiv(self, other)
 
@@ -475,13 +508,13 @@ class Apply(object):
         if self.o_len is not None and isinstance(idx, int):
             if idx >= self.o_len:
                 #  -- this IndexError is essential for supporting
-                #     tuple-unpacking syntax or list coersion of self.
+                #     tuple-unpacking syntax or list coercion of self.
                 raise IndexError()
         return scope.getitem(self, idx)
 
     def __len__(self):
         if self.o_len is None:
-            raise TypeError('len of pyll.Apply either undefined or unknown')
+            raise TypeError("len of pyll.Apply either undefined or unknown")
         return self.o_len
 
     def __call__(self, *args, **kwargs):
@@ -490,12 +523,9 @@ class Apply(object):
 
 def apply(name, *args, **kwargs):
     pos_args = [as_apply(a) for a in args]
-    named_args = [(k, as_apply(v)) for (k, v) in kwargs.items()]
+    named_args = [(k, as_apply(v)) for (k, v) in list(kwargs.items())]
     named_args.sort()
-    return Apply(name,
-            pos_args=pos_args,
-            named_args=named_args,
-            o_len=None)
+    return Apply(name, pos_args=pos_args, named_args=named_args, o_len=None)
 
 
 class Literal(Apply):
@@ -504,7 +534,7 @@ class Literal(Apply):
             o_len = len(obj)
         except TypeError:
             o_len = None
-        Apply.__init__(self, 'literal', [], {}, o_len, pure=True)
+        Apply.__init__(self, "literal", [], {}, o_len, pure=True)
         self._obj = obj
 
     def eval(self, memo=None):
@@ -526,26 +556,29 @@ class Literal(Apply):
         if memo is None:
             memo = {}
         if self in memo:
-            print >> ofile, lineno[0], ' ' * indent + memo[self]
+            print(lineno[0], " " * indent + memo[self], file=ofile)
         else:
             # TODO: set up a registry for this
             if isinstance(self._obj, np.ndarray):
-                msg = 'Literal{np.ndarray,shape=%s,min=%f,max=%f}' % (
-                        self._obj.shape, self._obj.min(), self._obj.max())
+                msg = "Literal{{np.ndarray,shape={},min={:f},max={:f}}}".format(
+                    self._obj.shape,
+                    self._obj.min(),
+                    self._obj.max(),
+                )
             else:
-                msg = 'Literal{%s}' % str(self._obj)
-            memo[self] = '%s  [line:%i]' % (msg, lineno[0])
-            print >> ofile, lineno[0], ' ' * indent + msg
+                msg = "Literal{%s}" % str(self._obj)
+            memo[self] = "%s  [line:%i]" % (msg, lineno[0])
+            print(lineno[0], " " * indent + msg, file=ofile)
         lineno[0] += 1
 
     def replace_input(self, old_node, new_node):
         return []
 
-    def clone_from_inputs(self, inputs, o_len='same'):
+    def clone_from_inputs(self, inputs, o_len="same"):
         return self.__class__(self._obj)
 
 
-class Lambda(object):
+class Lambda:
 
     # XXX: Extend Lambda objects to have a list of exception clauses.
     #      If the code of the expr() throws an error, these clauses convert
@@ -554,28 +587,23 @@ class Lambda(object):
     def __init__(self, name, params, expr):
         self.__name__ = name  # like a python function
         self.params = params  # list of (name, symbol[, default_value]) tuples
-        self.expr = expr      # pyll graph defining this Lambda
+        self.expr = expr  # pyll graph defining this Lambda
 
     def __call__(self, *args, **kwargs):
         # -- return `expr` cloned from given args and kwargs
         if len(args) > len(self.params):
-            raise TypeError('too many arguments')
+            raise TypeError("too many arguments")
         memo = {}
         for arg, param in zip(args, self.params):
-            #print 'applying with arg', param, arg
+            # print('applying with arg', param, arg)
             memo[param[1]] = as_apply(arg)
         if len(args) != len(self.params) or kwargs:
-            raise NotImplementedError('named / default arguments',
-                    (args, self.params))
+            raise NotImplementedError("named / default arguments", (args, self.params))
         rval = clone(self.expr, memo)
-        #print 'BEFORE'
-        #print self.expr
-        #print 'AFTER'
-        #print rval
         return rval
 
 
-class UndefinedValue(object):
+class UndefinedValue:
     pass
 
 
@@ -589,7 +617,7 @@ p4 = Literal(UndefinedValue)
 
 @scope.define
 def call(fn, args=(), kwargs={}):
-    """ call fn with given args and kwargs.
+    """call fn with given args and kwargs.
 
     This is used to represent Apply.__call__
     """
@@ -631,9 +659,8 @@ def partial(name, *args, **kwargs):
     my_id = len(scope._impls)
     # -- create a function with this name
     #    the name is the string used index into scope._impls
-    temp_name = 'partial_%s_id%i' % (name, my_id)
-    l = Lambda(temp_name, [('x', p0)],
-            expr=apply(name, *(args + (p0,)), **kwargs))
+    temp_name = "partial_%s_id%i" % (name, my_id)
+    l = Lambda(temp_name, [("x", p0)], expr=apply(name, *(args + (p0,)), **kwargs))
     scope.define(l)
     # assert that the next partial will get a different id
     # XXX; THIS ASSUMES THAT SCOPE ONLY GROWS
@@ -646,14 +673,17 @@ def dfs(aa, seq=None, seqset=None):
     if seq is None:
         assert seqset is None
         seq = []
-        seqset = set()
+        seqset = {}
     # -- seqset is the set of all nodes we have seen (which may be still on
     #    the stack)
+    #    N.B. it used to be a stack, but now it's a dict mapping to inputs
+    #    because that's an optimization saving us from having to call inputs
+    #    so often.
     if aa in seqset:
         return
     assert isinstance(aa, Apply)
-    seqset.add(aa)
-    for ii in aa.inputs():
+    seqset[aa] = aa.inputs()
+    for ii in seqset[aa]:
         dfs(ii, seq, seqset)
     seq.append(aa)
     return seq
@@ -661,7 +691,7 @@ def dfs(aa, seq=None, seqset=None):
 
 def toposort(expr):
     """
-    Return apply nodes of `expr` sub-tree in topological order.
+    Return apply nodes of `expr` sub-tree as a list in topological order.
 
     Raises networkx.NetworkXUnfeasible if subtree contains cycle.
 
@@ -669,7 +699,7 @@ def toposort(expr):
     G = nx.DiGraph()
     for node in dfs(expr):
         G.add_edges_from([(n_in, node) for n_in in node.inputs()])
-    order = nx.topological_sort(G)
+    order = list(nx.topological_sort(G))
     assert order[-1] == expr
     return order
 
@@ -693,7 +723,10 @@ def clone_merge(expr, memo=None, merge_literals=False):
     # -- args are somewhat slow to construct, so cache them out front
     #    XXX node.arg does not always work (builtins, weird co_flags)
     node_args = [(node.pos_args, node.named_args) for node in nodes]
-    del node
+    try:
+        del node
+    except:
+        pass
     for ii, node_ii in enumerate(nodes):
         if node_ii in memo:
             continue
@@ -703,7 +736,7 @@ def clone_merge(expr, memo=None, merge_literals=False):
                 node_jj = nodes[jj]
                 if node_ii.name != node_jj.name:
                     continue
-                if node_ii.name == 'literal':
+                if node_ii.name == "literal":
                     if not merge_literals:
                         continue
                     if node_ii._obj != node_jj._obj:
@@ -711,8 +744,7 @@ def clone_merge(expr, memo=None, merge_literals=False):
                 else:
                     if node_args[ii] != node_args[jj]:
                         continue
-                logger.debug('clone_merge %s %i <- %i' % (
-                    node_ii.name, jj, ii))
+                logger.debug("clone_merge %s %i <- %i" % (node_ii.name, jj, ii))
                 new_ii = node_jj
                 break
         if new_ii is None:
@@ -727,16 +759,19 @@ def clone_merge(expr, memo=None, merge_literals=False):
 ##############################################################################
 
 
-class GarbageCollected(object):
-    '''Placeholder representing a garbage-collected value '''
+class GarbageCollected:
+    """Placeholder representing a garbage-collected value"""
 
 
-def rec_eval(expr, deepcopy_inputs=False, memo=None,
-        max_program_len=None,
-        memo_gc=True,
-        print_trace=False,
-        print_node_on_error=True,
-        ):
+def rec_eval(
+    expr,
+    deepcopy_inputs=False,
+    memo=None,
+    max_program_len=None,
+    memo_gc=True,
+    print_trace=False,
+    print_node_on_error=True,
+):
     """
     expr - pyll Apply instance to be evaluated
 
@@ -758,7 +793,7 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
     if deepcopy_inputs not in (0, 1, False, True):
         # -- I've been calling rec_eval(expr, memo) by accident a few times
         #    this error would have been appreciated.
-        raise ValueError('deepcopy_inputs should be bool', deepcopy_inputs)
+        raise ValueError("deepcopy_inputs should be bool", deepcopy_inputs)
 
     node = as_apply(expr)
     topnode = node
@@ -768,6 +803,13 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
     else:
         memo = dict(memo)
 
+    # -- hack for speed
+    #    since the inputs are constant during rec_eval
+    #    but not constant in general
+    node_inputs = {}
+    node_list = []
+    dfs(node, node_list, seqset=node_inputs)
+
     # TODO: optimize dfs to not recurse past the items in memo
     #       this is especially important for evaluating Lambdas
     #       which cause rec_eval to recurse
@@ -776,38 +818,40 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
     #      so that this iteration may be an incomplete
     if memo_gc:
         clients = {}
-        for aa in dfs(node):
+        for aa in node_list:
             clients.setdefault(aa, set())
-            for ii in aa.inputs():
+            for ii in node_inputs[aa]:
                 clients.setdefault(ii, set()).add(aa)
+
         def set_memo(k, v):
             assert v is not GarbageCollected
             memo[k] = v
-            for ii in k.inputs():
+            for ii in node_inputs[k]:
                 # -- if all clients of ii are already in the memo
                 #    then we can free memo[ii] by replacing it
                 #    with a dummy symbol
                 if all(iic in memo for iic in clients[ii]):
-                    #print 'collecting', ii
                     memo[ii] = GarbageCollected
+
     else:
+
         def set_memo(k, v):
             memo[k] = v
 
     todo = deque([topnode])
     while todo:
         if len(todo) > max_program_len:
-            raise RuntimeError('Probably infinite loop in document')
+            raise RuntimeError("Probably infinite loop in document")
         node = todo.pop()
         if print_trace:
-            print 'rec_eval:print_trace', len(todo), node.name
+            print("rec_eval:print_trace", len(todo), node.name)
 
         if node in memo:
             # -- we've already computed this, move on.
             continue
 
         # -- different kinds of nodes are treated differently:
-        if node.name == 'switch':
+        if node.name == "switch":
             # -- switch is the conditional evaluation node
             switch_i_var = node.pos_args[0]
             if switch_i_var in memo:
@@ -815,11 +859,10 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
                 try:
                     int(switch_i)
                 except:
-                    raise TypeError('switch argument was', switch_i)
+                    raise TypeError("switch argument was", switch_i)
                 if switch_i != int(switch_i) or switch_i < 0:
-                    raise ValueError('switch pos must be positive int',
-                            switch_i)
-                rval_var = node.pos_args[switch_i + 1]
+                    raise ValueError("switch pos must be positive int", switch_i)
+                rval_var = node.pos_args[int(switch_i) + 1]
                 if rval_var in memo:
                     set_memo(node, memo[rval_var])
                     continue
@@ -833,7 +876,7 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
             continue
         else:
             # -- normal instruction-type nodes have inputs
-            waiting_on = [v for v in node.inputs() if v not in memo]
+            waiting_on = [v for v in node_inputs[node] if v not in memo]
 
         if waiting_on:
             # -- Necessary inputs have yet to be evaluated.
@@ -845,11 +888,10 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
             # -- not waiting on anything;
             #    this instruction can be evaluated.
             args = _args = [memo[v] for v in node.pos_args]
-            kwargs = _kwargs = dict([(k, memo[v])
-                for (k, v) in node.named_args])
+            kwargs = _kwargs = {k: memo[v] for (k, v) in node.named_args}
 
             if memo_gc:
-                for aa in args + kwargs.values():
+                for aa in args + list(kwargs.values()):
                     assert aa is not GarbageCollected
 
             if deepcopy_inputs:
@@ -859,14 +901,14 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
             try:
                 rval = scope._impls[node.name](*args, **kwargs)
 
-            except Exception, e:
+            except Exception as e:
                 if print_node_on_error:
-                    print '=' * 80
-                    print 'ERROR in rec_eval'
-                    print 'EXCEPTION', type(e), str(e)
-                    print 'NODE'
-                    print node  # -- typically a multi-line string
-                    print '=' * 80
+                    print("=" * 80)
+                    print("ERROR in rec_eval")
+                    print("EXCEPTION", type(e), str(e))
+                    print("NODE")
+                    print(node)  # -- typically a multi-line string
+                    print("=" * 80)
                 raise
 
             if isinstance(rval, Apply):
@@ -875,8 +917,7 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
                 #
                 # XXX: consider if it is desirable, efficient, buggy
                 #      etc. to keep using the same memo dictionary
-                foo = rec_eval(rval, deepcopy_inputs, memo,
-                        memo_gc=memo_gc)
+                foo = rec_eval(rval, deepcopy_inputs, memo, memo_gc=memo_gc)
                 set_memo(node, foo)
             else:
                 set_memo(node, rval)
@@ -887,74 +928,37 @@ def rec_eval(expr, deepcopy_inputs=False, memo=None,
 ############################################################################
 ############################################################################
 
+
 @scope.define_pure
 def pos_args(*args):
     return args
 
 
-@scope.define_pure
-def getitem(obj, idx):
-    return obj[idx]
-
-
 @scope.define_pure
 def identity(obj):
     return obj
 
 
-@scope.define_pure
-def add(a, b):
-    return a + b
-
-
-@scope.define_pure
-def sub(a, b):
-    return a - b
-
-
-@scope.define_pure
-def neg(a):
-    return -a
-
-
-@scope.define_pure
-def mul(a, b):
-    return a * b
-
-
-@scope.define_pure
-def div(a, b):
-    return a / b
-
-
-@scope.define_pure
-def floordiv(a, b):
-    return a // b
-
-
-@scope.define_pure
-def eq(a, b):
-    return a == b
-
-
-@scope.define_pure
-def gt(a, b):
-    return a > b
-
-
-@scope.define_pure
-def ge(a, b):
-    return a >= b
-
-
-@scope.define_pure
-def lt(a, b):
-    return a < b
-
-
-@scope.define_pure
-def le(a, b):
-    return a <= b
+# -- We used to define these as Python functions in this file, but the operator
+#    module already provides them, is slightly more efficient about it. Since
+#    searchspaces uses the same convention, we can more easily map graphs back
+#    and forth and reduce the amount of code in both codebases.
+scope.define_pure(operator.getitem)
+scope.define_pure(operator.add)
+scope.define_pure(operator.sub)
+scope.define_pure(operator.mul)
+try:
+    scope.define_pure(operator.div)
+except AttributeError:
+    pass  # No more operator.div in Python3, but truediv also exists since Python2.2
+scope.define_pure(operator.truediv)
+scope.define_pure(operator.floordiv)
+scope.define_pure(operator.neg)
+scope.define_pure(operator.eq)
+scope.define_pure(operator.lt)
+scope.define_pure(operator.le)
+scope.define_pure(operator.gt)
+scope.define_pure(operator.ge)
 
 
 @scope.define_pure
@@ -1036,33 +1040,15 @@ def str_join(s, seq):
     return s.join(seq)
 
 
-def _bincount_slow(x, weights=None, minlength=None):
-    """backport of np.bincount post numpy 1.6
-    """
-    if weights is not None:
-        raise NotImplementedError()
-    if minlength is None:
-        rlen = np.max(x) + 1
-    else:
-        rlen = max(np.max(x) + 1, minlength)
-    rval = np.zeros(rlen, dtype='int')
-    for xi in np.asarray(x).flatten():
-        rval[xi] += 1
-    return rval
-
-
 @scope.define_pure
-def bincount(x, weights=None, minlength=None):
-    if np_versions[0] == 1 and np_versions[1] < 6:
-        # -- np.bincount doesn't have minlength arg
-        return _bincount_slow(x, weights, minlength)
-    else:
-        if np.asarray(x).size:
-            return np.bincount(x, weights, minlength)
-        else:
-            # -- currently numpy rejects this case,
-            #    but it seems sensible enough to me.
-            return np.zeros(minlength, dtype='int')
+def bincount(x, offset=0, weights=None, minlength=None, p=None):
+    y = np.asarray(x, dtype="int")
+    # hack for pchoice, p is passed as [ np.repeat(p, obs.size) ],
+    # so scope.len(p) gives incorrect #dimensions, need to get just the first one
+    if p is not None and p.ndim == 2:
+        assert np.all(p == p[0])
+        minlength = len(p[0])
+    return np.bincount(y - offset, weights, minlength)
 
 
 @scope.define_pure
@@ -1099,7 +1085,7 @@ def switch(pos, *args):
     # switch is an unusual expression, in that it affects control flow
     # when executed with rec_eval. args are not all evaluated, only
     # args[pos] is evaluated.
-    ##raise RuntimeError('switch is not meant to be evaluated')
+    # raise RuntimeError('switch is not meant to be evaluated')
     #
     # .. However, in quick-evaluation schemes it is handy that this be defined
     # as follows:
@@ -1109,10 +1095,11 @@ def switch(pos, *args):
 def _kwswitch(kw, **kwargs):
     """conditional evaluation according to string value"""
     # Get the index of the string in kwargs to use switch
-    keys, values = zip(*sorted(kwargs.iteritems()))
-    match_idx = scope.call_method_pure(keys, 'index', kw)
+    keys, values = list(zip(*sorted(kwargs.items())))
+    match_idx = scope.call_method_pure(keys, "index", kw)
     return scope.switch(match_idx, *values)
 
+
 scope.kwswitch = _kwswitch
 
 
@@ -1128,6 +1115,7 @@ def curtime(obj):
 
 @scope.define
 def pdb_settrace(obj):
-    import pdb; pdb.set_trace()
-    return obj
+    import pdb
 
+    pdb.set_trace()
+    return obj
