@@ -1,15 +1,17 @@
 """
 Constructs for annotating base graphs.
 """
+from past.utils import old_div
 import sys
 import numpy as np
-
-from .base import scope, as_apply, dfs, Apply, rec_eval, clone
+from .base import scope, as_apply, dfs, rec_eval, clone
 
 ################################################################################
 ################################################################################
+
+
 def ERR(msg):
-    print >> sys.stderr, msg
+    print(msg, file=sys.stderr)
 
 
 implicit_stochastic_symbols = set()
@@ -22,11 +24,12 @@ def implicit_stochastic(f):
 
 @scope.define
 def rng_from_seed(seed):
-    return np.random.RandomState(seed)
+    return np.random.default_rng(seed)
 
 
 # -- UNIFORM
 
+
 @implicit_stochastic
 @scope.define
 def uniform(low, high, rng=None, size=()):
@@ -44,18 +47,19 @@ def loguniform(low, high, rng=None, size=()):
 @scope.define
 def quniform(low, high, q, rng=None, size=()):
     draw = rng.uniform(low, high, size=size)
-    return np.round(draw/q) * q
+    return np.round(old_div(draw, q)) * q
 
 
 @implicit_stochastic
 @scope.define
 def qloguniform(low, high, q, rng=None, size=()):
     draw = np.exp(rng.uniform(low, high, size=size))
-    return np.round(draw/q) * q
+    return np.round(old_div(draw, q)) * q
 
 
 # -- NORMAL
 
+
 @implicit_stochastic
 @scope.define
 def normal(mu, sigma, rng=None, size=()):
@@ -66,7 +70,7 @@ def normal(mu, sigma, rng=None, size=()):
 @scope.define
 def qnormal(mu, sigma, q, rng=None, size=()):
     draw = rng.normal(mu, sigma, size=size)
-    return np.round(draw/q) * q
+    return np.round(old_div(draw, q)) * q
 
 
 @implicit_stochastic
@@ -80,7 +84,7 @@ def lognormal(mu, sigma, rng=None, size=()):
 @scope.define
 def qlognormal(mu, sigma, q, rng=None, size=()):
     draw = np.exp(rng.normal(mu, sigma, size=size))
-    return np.round(draw/q) * q
+    return np.round(old_div(draw, q)) * q
 
 
 # -- CATEGORICAL
@@ -88,22 +92,30 @@ def qlognormal(mu, sigma, q, rng=None, size=()):
 
 @implicit_stochastic
 @scope.define
-def randint(upper, rng=None, size=()):
-    # this is tricky because numpy doesn't support
-    # upper being a list of len size[0]
-    if isinstance(upper, (list, tuple)):
-        if isinstance(size, int):
-            assert len(upper) == size
-            return np.asarray([rng.randint(uu) for uu in upper])
-        elif len(size) == 1:
-            assert len(upper) == size[0]
-            return np.asarray([rng.randint(uu) for uu in upper])
-    return rng.randint(upper, size=size)
+def randint(low, high=None, rng=None, size=()):
+    """
+    See np.random.randint documentation.
+    rng = random number generator, typically equals np.random.Generator
+    """
+    return rng.integers(low, high, size)
 
 
 @implicit_stochastic
 @scope.define
-def categorical(p, upper=None, rng=None, size=()):
+def randint_via_categorical(p, rng=None, size=()):
+    """
+    Only used in tpe because of the chaotic API based on names.
+    # ideally we would just use randint above, but to use priors this is a wrapper of
+    categorical
+    rng = random number generator, typically equals np.random.Generator
+    """
+
+    return scope.categorical(p, rng, size)
+
+
+@implicit_stochastic
+@scope.define
+def categorical(p, rng=None, size=()):
     """Draws i with probability p[i]"""
     if len(p) == 1 and isinstance(p[0], np.ndarray):
         p = p[0]
@@ -131,10 +143,11 @@ def categorical(p, upper=None, rng=None, size=()):
         return rval
     elif p.ndim == 2:
         n_draws_, n_choices = p.shape
-        n_draws, = size
+        (n_draws,) = size
         assert n_draws == n_draws_
-        rval = [np.where(rng.multinomial(pvals=p[ii], n=1))[0][0]
-                                for ii in xrange(n_draws)]
+        rval = [
+            np.where(rng.multinomial(pvals=p[ii], n=1))[0][0] for ii in range(n_draws)
+        ]
         rval = np.asarray(rval)
         rval.shape = size
         return rval
@@ -144,12 +157,16 @@ def categorical(p, upper=None, rng=None, size=()):
 
 def choice(args):
     return scope.one_of(*args)
+
+
 scope.choice = choice
 
 
 def one_of(*args):
     ii = scope.randint(len(args))
     return scope.switch(ii, *args)
+
+
 scope.one_of = one_of
 
 
@@ -160,11 +177,16 @@ def recursive_set_rng_kwarg(expr, rng=None):
     uniform(0, 1) -> uniform(0, 1, rng=rng)
     """
     if rng is None:
-        rng = np.random.RandomState()
+        rng = np.random.default_rng()
     lrng = as_apply(rng)
     for node in dfs(expr):
         if node.name in implicit_stochastic_symbols:
-            node.named_args.append(('rng', lrng))
+            for ii, (name, arg) in enumerate(list(node.named_args)):
+                if name == "rng":
+                    node.named_args[ii] = ("rng", lrng)
+                    break
+            else:
+                node.named_args.append(("rng", lrng))
     return expr
 
 
@@ -173,14 +195,14 @@ def sample(expr, rng=None, **kwargs):
     Parameters:
     expr - a pyll expression to be evaluated
 
-    rng - a np.random.RandomState instance
-          default: `np.random.RandomState()`
-          
+    rng - a np.random.Generator instance
+          default: `np.random.default_rng()`
+
     **kwargs - optional arguments passed along to
                `hyperopt.pyll.rec_eval`
 
     """
     if rng is None:
-        rng = np.random.RandomState()
+        rng = np.random.default_rng()
     foo = recursive_set_rng_kwarg(clone(as_apply(expr)), as_apply(rng))
     return rec_eval(foo, **kwargs)
